#region License
/* Copyright 2005, 2013 James F. Bellinger
 * 
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redestribute it
 * freely, subject to the following restrictions:
 * 
 * 1. The origin of this software must not be misrepresented; you must
 * not claim that you wrote the original software. If you use this software
 * in a product, an acknowledgement in the product documentation would be
 * appreciated but is not required.
 * 
 * 2. Altered source versions must be plainly marked as such, and must
 * not be misrepresented as being the original software.
 * 
 * 3. This notice may not be removed or altered from any source
 * distribution. */
#endregion

using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Net.Security;
using System.Net.Sockets;
using System.Reflection;
using System.Security.Authentication;
using System.Windows.Forms;
using Chatter.Logic;
using Chatter.Plugins;
using Chatter.Plugins.Events;
using Chatter.Plugins.Server;
using NetIrc2;
using NetIrc2.Events;
using NetIrc2.Parsing;
using SecondLanguage;

namespace Chatter
{
	public sealed class ChatterMDI: Form,
        IPluginServer,
		IChatController,
		IChatEvents,
		ISettings
	{
        static Translator _ = Translator.Default;

        // IPluginServer implementation.
        object IPluginServer.GetInterface(Type type)
        {
            return type == typeof(IChatController) || type == typeof(IChatEvents) || type == typeof(ISettings)
                ? this : null;
        }

		// ISettings implementation.
		string ISettings.ProgramDirectory
		{
            get { return Directories.ProgramDirectory; }
		}

		string ISettings.PluginDirectory
		{
            get { return Directories.PluginDirectory; }
		}

		object ISettings.Load(Type objectType)
		{
			string plugin = Assembly.GetCallingAssembly().GetName().Name;

            object settings;
            return SettingsIO.TryLoad(Directories.GetPluginSettingsFilename(plugin), objectType, out settings)
                ? settings : null;
		}

		void ISettings.Save(object settings)
		{
			string plugin = Assembly.GetCallingAssembly().GetName().Name;

            SettingsIO.TrySave(Directories.GetPluginSettingsFilename(plugin), settings);
		}

		// IChatController implementation.
		string IChatController.ActiveChannel
		{
			get
			{
                var channel = ActiveMdiChild as Channel;
                return channel != null && !channel.IsStatusWindow() ? channel.Text : null;
			}
		}

		string IChatController.ApplicationName
			{ get { return "Chatter"; } }

		bool IChatController.IsConnected
			{ get { return ircClient != null && ircClient.IsConnected; } }

		public string Nickname
			{ get { return trueNickname; } }

		Version IChatController.Version
			{ get { return Environment.Version; } }

		void IChatController.Message(string recipient, MessageType messageType, string message)
		{
			if (ircClient != null && message != null)
			{
				switch(messageType)
				{
					case MessageType.Action:
						ircClient.ChatAction(recipient, message); break;

					case MessageType.Message:
						ircClient.Message(recipient, message); break;

					case MessageType.Notice:
						ircClient.Notice(recipient, message); break;
				}

				ChannelMessage(Color.DarkBlue, recipient, "Plugin " +
					messageType.ToString() + " [" + recipient + "] -> " +
					message, true);

                InvokePluginMessage(Nickname, recipient, message, messageType,
                                    MessageFlags.Local | MessageFlags.Plugin);
			}
		}

		void IChatController.Join(string channelName)
		{
			if (ircClient != null) { ircClient.Join(channelName); }
		}

		void IChatController.Leave(string channelName)
		{
			if (ircClient != null) { ircClient.Leave(channelName); }
		}

		bool IChatController.QueryUser(string queryMessage,
			string queryTitle, out string userResponse)
		{
			return InputDialog.QueryDialog(queryMessage, queryTitle, s => true, out userResponse)
				== DialogResult.OK;
		}

		// IChatEvents implementation.
		private event EventHandler<EntryEventArgs>   ierEntry;
		private event EventHandler<MessageEventArgs> ierMessage;

        event EventHandler<EntryEventArgs> IChatEvents.Entry
		{
            add { ierEntry += value; }
            remove { ierEntry -= value; }
		}

        event EventHandler<MessageEventArgs> IChatEvents.Message
		{
            add { ierMessage += value; }
            remove { ierMessage -= value; }
		}

		// ChatterMDI implementation.
		private System.Windows.Forms.MainMenu mnuMain;
		private System.Windows.Forms.MenuItem menuFile;
		private System.Windows.Forms.MenuItem mnuCommands;
		private System.Windows.Forms.MenuItem mnuCommandsJoin;
		private System.Windows.Forms.MenuItem mnuCommandsLeave;
		private System.Windows.Forms.MenuItem mnuHelp;
		private System.Windows.Forms.MenuItem mnuHelpAbout;
		private System.Windows.Forms.MenuItem mnuFileLogin;
		private System.Windows.Forms.MenuItem mnuFileBar1;
		private System.Windows.Forms.MenuItem mnuFileExit;
        private System.Windows.Forms.MenuItem menuCommandsList;
        private System.ComponentModel.IContainer components;

		private Plugins.Manager pluginManager;

		private string trueNickname;
		private FontList fontList;
		private Login windowLogin;
		private ChannelList chanList;
		private System.Windows.Forms.MenuItem mnuCommandsBar1;
		private System.Windows.Forms.MenuItem mnuCommandsFilters;
		private System.Windows.Forms.MenuItem mnuCommandsFiltersEliz;
		private System.Windows.Forms.MenuItem mnuCommandsFiltersPropEng;
		private System.Windows.Forms.MenuItem mnuCommandsFiltersShort;
		private System.Windows.Forms.MenuItem mnuCommandsBar2;
        private System.Windows.Forms.MenuItem mnuCommandsNickname;
		internal System.Windows.Forms.MenuItem mnuRobots;
		private System.Windows.Forms.MenuItem mnuWindow;
		private System.Windows.Forms.MenuItem mnuWindowTileH;
		private System.Windows.Forms.MenuItem mnuWindowTileV;
		private System.Windows.Forms.MenuItem mnuWindowCascade;
		private System.Windows.Forms.MenuItem mnuWindowArrange;
        private IrcClient ircClient;
        private IdentServer identServer;

		private string StatusWindowName = "**:" + _["Status"];

		internal ChatterMDI()
		{
			InitializeComponent();

			System.IO.Directory.SetCurrentDirectory
				(((ISettings)this).ProgramDirectory);

			fontList = new FontList();
			channels = new List<ChannelArray>();

			UpdateFonts(); ResetTrueName();
		}

		void ResetTrueName()
		{
			trueNickname = Settings.Instance.UserInfo.Nickname;
		}

		internal List<Chatter.Plugins.Manager.Plugin> PluginList
		{
			get { return pluginManager.Plugins; }
		}

		internal FontList Fonts
		{
			get { return fontList; }
		}

		internal IrcClient Connection
		{
			get { return ircClient; }
		}

		internal void UpdateFonts()
		{
            Fonts.Channels = new Font(Settings.Instance.Fonts.Channels.Name, Settings.Instance.Fonts.Channels.Size);
            Fonts.ChatText = new Font(Settings.Instance.Fonts.ChatText.Name, Settings.Instance.Fonts.ChatText.Size);
            Fonts.UserList = new Font(Settings.Instance.Fonts.UserList.Name, Settings.Instance.Fonts.UserList.Size);
		}

		internal bool ConnectTo(string Hostname, int Port, bool SSL)
		{
			Disconnect();

			if (Settings.Instance.Identd.Enabled)
			{
				try
				{
                    identServer.OperatingSystem = Settings.Instance.Identd.OperatingSystem;
                    identServer.UserID = Settings.Instance.Identd.UserID;
                    identServer.Start();
				}
				catch (SocketException)
				{
					// Couldn't start the Ident server.
					ChannelMessage(Color.DarkKhaki, null, "Ident server failed to start.", true);
				}
			}

			mnuFileLogin.Text = "&Log out";
			mnuCommands.Enabled = true;

			ResetTrueName();

            var options = new IrcClientConnectionOptions();
            options.Ssl = SSL;
            options.SslCertificateValidationCallback = (sender, certificate, chain, sslPolicyErrors) =>
                {
                    if (sslPolicyErrors == SslPolicyErrors.None) { return true; }

                    return MessageBox.Show(this,
                                           "There is a problem with the server's SSL certificate.\n\nContinue anyway?",
                                           "Certificate Error",
                                           MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.Yes;
                };
            options.SslHostname = Hostname;

            try
            {
                ircClient.Connect(Hostname, Port, options);
            }
            catch (AuthenticationException)
            {
                Disconnect();
                return false;
            }
            catch (IOException)
            {
                Disconnect();
                return false;
            }
            catch (SocketException)
            {
                Disconnect();
                return false;
            }

			return true;
		}

		internal void Disconnect()
		{
			identServer.Stop();

			ircClient.LogOut(); ircClient.Close();

			while (channels.Count > 0)
			{
				// This relies on the fact that the Channel forms
				// remove themselves from the list when they are
				// closed.
				((ChannelArray)channels[0]).Form.Close();
			}

			mnuFileLogin.Text = "&Log in";
			mnuCommands.Enabled = false;
		}

		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				if (components != null) 
				{
					components.Dispose();
				}
			}

			base.Dispose(disposing);
		}

		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
            this.components = new System.ComponentModel.Container();
            System.ComponentModel.ComponentResourceManager resources = new System.ComponentModel.ComponentResourceManager(typeof(ChatterMDI));
            this.mnuMain = new System.Windows.Forms.MainMenu(this.components);
            this.menuFile = new System.Windows.Forms.MenuItem();
            this.mnuFileLogin = new System.Windows.Forms.MenuItem();
            this.mnuFileBar1 = new System.Windows.Forms.MenuItem();
            this.mnuFileExit = new System.Windows.Forms.MenuItem();
            this.mnuCommands = new System.Windows.Forms.MenuItem();
            this.mnuCommandsJoin = new System.Windows.Forms.MenuItem();
            this.mnuCommandsLeave = new System.Windows.Forms.MenuItem();
            this.menuCommandsList = new System.Windows.Forms.MenuItem();
            this.mnuCommandsBar1 = new System.Windows.Forms.MenuItem();
            this.mnuCommandsNickname = new System.Windows.Forms.MenuItem();
            this.mnuCommandsBar2 = new System.Windows.Forms.MenuItem();
            this.mnuCommandsFilters = new System.Windows.Forms.MenuItem();
            this.mnuCommandsFiltersEliz = new System.Windows.Forms.MenuItem();
            this.mnuCommandsFiltersPropEng = new System.Windows.Forms.MenuItem();
            this.mnuCommandsFiltersShort = new System.Windows.Forms.MenuItem();
            this.mnuRobots = new System.Windows.Forms.MenuItem();
            this.mnuWindow = new System.Windows.Forms.MenuItem();
            this.mnuWindowCascade = new System.Windows.Forms.MenuItem();
            this.mnuWindowTileH = new System.Windows.Forms.MenuItem();
            this.mnuWindowTileV = new System.Windows.Forms.MenuItem();
            this.mnuWindowArrange = new System.Windows.Forms.MenuItem();
            this.mnuHelp = new System.Windows.Forms.MenuItem();
            this.mnuHelpAbout = new System.Windows.Forms.MenuItem();
            this.ircClient = new NetIrc2.IrcClient();
            this.identServer = new NetIrc2.IdentServer();
            this.SuspendLayout();
            // 
            // mnuMain
            // 
            this.mnuMain.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
            this.menuFile,
            this.mnuCommands,
            this.mnuRobots,
            this.mnuWindow,
            this.mnuHelp});
            // 
            // menuFile
            // 
            this.menuFile.Index = 0;
            this.menuFile.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
            this.mnuFileLogin,
            this.mnuFileBar1,
            this.mnuFileExit});
            this.menuFile.Text = "&File";
            // 
            // mnuFileLogin
            // 
            this.mnuFileLogin.Index = 0;
            this.mnuFileLogin.Text = "&Log in";
            this.mnuFileLogin.Click += new System.EventHandler(this.mnuFileLogin_Click);
            // 
            // mnuFileBar1
            // 
            this.mnuFileBar1.Index = 1;
            this.mnuFileBar1.Text = "-";
            // 
            // mnuFileExit
            // 
            this.mnuFileExit.Index = 2;
            this.mnuFileExit.Text = "E&xit";
            this.mnuFileExit.Click += new System.EventHandler(this.mnuFileExit_Click);
            // 
            // mnuCommands
            // 
            this.mnuCommands.Enabled = false;
            this.mnuCommands.Index = 1;
            this.mnuCommands.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
            this.mnuCommandsJoin,
            this.mnuCommandsLeave,
            this.menuCommandsList,
            this.mnuCommandsBar1,
            this.mnuCommandsNickname,
            this.mnuCommandsBar2,
            this.mnuCommandsFilters});
            this.mnuCommands.Text = "&Commands";
            // 
            // mnuCommandsJoin
            // 
            this.mnuCommandsJoin.Index = 0;
            this.mnuCommandsJoin.Text = "&Join channel";
            this.mnuCommandsJoin.Click += new System.EventHandler(this.mnuCommandsJoin_Click);
            // 
            // mnuCommandsLeave
            // 
            this.mnuCommandsLeave.Index = 1;
            this.mnuCommandsLeave.Text = "&Leave channel";
            this.mnuCommandsLeave.Click += new System.EventHandler(this.mnuCommandsLeave_Click);
            // 
            // menuCommandsList
            // 
            this.menuCommandsList.Index = 2;
            this.menuCommandsList.Text = "List &channels";
            this.menuCommandsList.Click += new System.EventHandler(this.menuCommandsList_Click);
            // 
            // mnuCommandsBar1
            // 
            this.mnuCommandsBar1.Index = 3;
            this.mnuCommandsBar1.Text = "-";
            // 
            // mnuCommandsNickname
            // 
            this.mnuCommandsNickname.Index = 4;
            this.mnuCommandsNickname.Text = "Change &nickname";
            this.mnuCommandsNickname.Click += new System.EventHandler(this.mnuCommandsNickname_Click);
            // 
            // mnuCommandsBar2
            // 
            this.mnuCommandsBar2.Index = 5;
            this.mnuCommandsBar2.Text = "-";
            // 
            // mnuCommandsFilters
            // 
            this.mnuCommandsFilters.Index = 6;
            this.mnuCommandsFilters.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
            this.mnuCommandsFiltersEliz,
            this.mnuCommandsFiltersPropEng,
            this.mnuCommandsFiltersShort});
            this.mnuCommandsFilters.Text = "Text &filters";
            // 
            // mnuCommandsFiltersEliz
            // 
            this.mnuCommandsFiltersEliz.Index = 0;
            this.mnuCommandsFiltersEliz.Text = "Elizabethan";
            this.mnuCommandsFiltersEliz.Click += new System.EventHandler(this.mnuCommandsFiltersEliz_Click);
            // 
            // mnuCommandsFiltersPropEng
            // 
            this.mnuCommandsFiltersPropEng.Index = 1;
            this.mnuCommandsFiltersPropEng.Text = "Proper English";
            this.mnuCommandsFiltersPropEng.Click += new System.EventHandler(this.mnuCommandsFiltersPropEng_Click);
            // 
            // mnuCommandsFiltersShort
            // 
            this.mnuCommandsFiltersShort.Index = 2;
            this.mnuCommandsFiltersShort.Text = "Shorthand";
            this.mnuCommandsFiltersShort.Click += new System.EventHandler(this.mnuCommandsFiltersShort_Click);
            // 
            // mnuRobots
            // 
            this.mnuRobots.Enabled = false;
            this.mnuRobots.Index = 2;
            this.mnuRobots.Text = "&Plugins";
            // 
            // mnuWindow
            // 
            this.mnuWindow.Index = 3;
            this.mnuWindow.MdiList = true;
            this.mnuWindow.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
            this.mnuWindowCascade,
            this.mnuWindowTileH,
            this.mnuWindowTileV,
            this.mnuWindowArrange});
            this.mnuWindow.Text = "&Window";
            // 
            // mnuWindowCascade
            // 
            this.mnuWindowCascade.Index = 0;
            this.mnuWindowCascade.Text = "&Cascade";
            this.mnuWindowCascade.Click += new System.EventHandler(this.mnuWindowCascade_Click);
            // 
            // mnuWindowTileH
            // 
            this.mnuWindowTileH.Index = 1;
            this.mnuWindowTileH.Text = "Tile &Horizontally";
            this.mnuWindowTileH.Click += new System.EventHandler(this.mnuWindowTileH_Click);
            // 
            // mnuWindowTileV
            // 
            this.mnuWindowTileV.Index = 2;
            this.mnuWindowTileV.Text = "&Tile Vertically";
            this.mnuWindowTileV.Click += new System.EventHandler(this.mnuWindowTileV_Click);
            // 
            // mnuWindowArrange
            // 
            this.mnuWindowArrange.Index = 3;
            this.mnuWindowArrange.Text = "&Arrange Icons";
            this.mnuWindowArrange.Click += new System.EventHandler(this.mnuWindowArrange_Click);
            // 
            // mnuHelp
            // 
            this.mnuHelp.Index = 4;
            this.mnuHelp.MenuItems.AddRange(new System.Windows.Forms.MenuItem[] {
            this.mnuHelpAbout});
            this.mnuHelp.Text = "&Help";
            // 
            // mnuHelpAbout
            // 
            this.mnuHelpAbout.Index = 0;
            this.mnuHelpAbout.Text = "&About this program";
            this.mnuHelpAbout.Click += new System.EventHandler(this.mnuHelpAbout_Click);
            // 
            // ircClient
            // 
            this.ircClient.ClientVersion = ((NetIrc2.IrcString)(resources.GetObject("ircClient.ClientVersion")));
            this.ircClient.GotChannelListBegin += new System.EventHandler(this.Chatter_OnChannelListBegin);
            this.ircClient.GotChannelListEntry += new System.EventHandler<NetIrc2.Events.ChannelListEntryEventArgs>(this.Chatter_OnChannelListEntry);
            this.ircClient.GotChannelListEnd += new System.EventHandler(this.Chatter_OnChannelListEnd);
            this.ircClient.GotChatAction += new System.EventHandler<NetIrc2.Events.ChatMessageEventArgs>(this.Chatter_OnChatAction);
            this.ircClient.GotIrcError += new System.EventHandler<NetIrc2.Events.IrcErrorEventArgs>(this.Chatter_OnError);
            this.ircClient.GotJoinChannel += new System.EventHandler<NetIrc2.Events.JoinLeaveEventArgs>(this.Chatter_OnJoin);
            this.ircClient.GotLeaveChannel += new System.EventHandler<NetIrc2.Events.JoinLeaveEventArgs>(this.Chatter_OnLeave);
            this.ircClient.GotMessage += new System.EventHandler<NetIrc2.Events.ChatMessageEventArgs>(this.Chatter_OnMessage);
            this.ircClient.GotMode += new System.EventHandler<NetIrc2.Events.ModeEventArgs>(this.Chatter_OnMode);
            this.ircClient.GotMotdText += new System.EventHandler<NetIrc2.Events.SimpleMessageEventArgs>(this.Chatter_OnMOTDText);
            this.ircClient.GotNameChange += new System.EventHandler<NetIrc2.Events.NameChangeEventArgs>(this.Chatter_OnNameChange);
            this.ircClient.GotNameListReply += new System.EventHandler<NetIrc2.Events.NameListReplyEventArgs>(this.Chatter_OnNameListReply);
            this.ircClient.GotNotice += new System.EventHandler<NetIrc2.Events.ChatMessageEventArgs>(this.Chatter_OnNotice);
            this.ircClient.GotPingReply += new System.EventHandler<NetIrc2.Events.PingReplyEventArgs>(this.Chatter_OnPingReply);
            this.ircClient.GotUserKicked += new System.EventHandler<NetIrc2.Events.KickEventArgs>(this.Chatter_OnKick);
            this.ircClient.GotUserQuit += new System.EventHandler<NetIrc2.Events.QuitEventArgs>(this.Chatter_OnQuit);
            // 
            // ChatterMDI
            // 
            this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
            this.ClientSize = new System.Drawing.Size(660, 578);
            this.IsMdiContainer = true;
            this.Menu = this.mnuMain;
            this.Name = "ChatterMDI";
            this.StartPosition = System.Windows.Forms.FormStartPosition.CenterScreen;
            this.Text = "Chatter";
            this.Load += new System.EventHandler(this.Chatter_Load);
            this.ResumeLayout(false);

		}
		#endregion

		private void Chatter_Load(object sender, System.EventArgs e)
		{
			Closed += new EventHandler(Chatter_Closed);

			mnuCommandsFiltersEliz.Checked    = Settings.Instance.Language.Filters.Elizabethan;
            mnuCommandsFiltersPropEng.Checked = Settings.Instance.Language.Filters.ProperEnglish;
            mnuCommandsFiltersShort.Checked   = Settings.Instance.Language.Filters.Shorthand;

			pluginManager = new Plugins.Manager(this);

            mnuFileLogin_Click(sender, e);
		}

		private void Chatter_Closed(object sender, System.EventArgs e)
		{
			pluginManager.Dispose(); pluginManager = null;
			Disconnect();

			Application.Exit();
		}

		private void mnuFileLogin_Click(object sender, System.EventArgs e)
		{
			if (mnuFileLogin.Text == "&Log out")
			{
				Cursor.Current = Cursors.WaitCursor;
				Disconnect();
				Cursor.Current = Cursors.Default;
			}
			else
			{
				if (windowLogin != null)
				{
					if (windowLogin.Visible) { return; }

					windowLogin.Dispose();
				}

				windowLogin = new Login(this);
				windowLogin.MdiParent = this;
				windowLogin.Show();
			}
		}

		private void mnuFileExit_Click(object sender, System.EventArgs e)
		{
			Close();
		}

		private void mnuCommandsJoin_Click(object sender, System.EventArgs e)
		{
			string channel;
			DialogResult result = InputDialog.QueryDialog
				("Join which channel?", "Join", s => IrcValidation.IsChannelName(s), out channel);

			if (result == DialogResult.OK)
			{
				JoinChannel(channel.Trim());
			}
		}

		private class ChannelArray
		{
			public string  Channel;
			public Channel Form;
		}

		List<ChannelArray> channels;

		ChannelArray FindChannelWindow(string channel)
		{
			channel = channel.Trim().ToUpper();

			foreach (ChannelArray ca in channels)
			{
				if (ca.Channel.Trim().ToUpper() == channel)
				{
					return ca;
				}
			}

			return null;
		}

		internal void JoinChannel(string channel)
		{
			ChannelArray ca = FindChannelWindow(channel);

			if (ca == null) { ircClient.Join(channel); }
		}

		internal void LeaveChannel(string channel)
		{
			ChannelArray ca = FindChannelWindow(channel);

			if (ca != null)
			{
				if (ircClient != null && channel.StartsWith("#"))
					{ ircClient.Leave(channel); }

				channels.Remove(ca);
			}
		}

		internal void FormChatWindow(string Title)
		{
			if (FindChannelWindow(Title) == null)
				{ CreateChatWindow(Title); }
		}

		private void ChannelListMessage(IrcString channelName, IrcString[] users, bool doRemove)
		{
            ChannelArray ca = FindChannelWindow(channelName);

            if (doRemove)
            {
                if (ca != null)
                {
                    foreach (string user in users) { ca.Form.UserRemove(user); }
                }
            }
            else
            {
                if (ca == null)
                {
                    if (channelName.StartsWith("#")) { return; }

                    ca = CreateChatWindow(channelName);
                }

                ca.Form.UserAdd(users.Select(user => (string)user).ToArray());
            }
		}

		ChannelArray CreateChatWindow(string channelName)
		{
			ChannelArray ca = new ChannelArray();
			ca.Channel = channelName;
			ca.Form = new Channel(this, channelName);
			ca.Form.MdiParent = this;
			ca.Form.Show(); channels.Add(ca);
			
			return ca;
		}

		private void ChannelMessage(Color messageColor, IrcString channelName,
			IrcString messageText, bool isNotice)
		{
            ChannelArray ca = FindChannelWindow(channelName == null
                ? StatusWindowName : (string)channelName);

            if (ca == null)
            {
                if (isNotice)
                {
                    ChannelMessage(messageColor, StatusWindowName, messageText, false);

                    return;
                }
                else
                {
                    if (channelName.StartsWith("#"))
                    {
                        if (messageText == null)
                        {
                            CreateChatWindow(channelName);
                        }

                        return;
                    }
                    else
                    {
                        ca = CreateChatWindow(channelName);
                    }
                }
            }

            if (messageText != null)
            {
                ca.Form.ChatEntry(messageColor, messageText);
            }
        }

		private void Chatter_OnJoin(object o, JoinLeaveEventArgs e)
		{
			foreach (IrcString channel in e.GetChannelList())
			{
				ChannelMessage(Color.Green, channel, _["{0} joined {1}", e.Identity.Nickname, channel], false);
				ChannelListMessage(channel, new[] { e.Identity.Nickname }, false);

				if (ierEntry != null)
				{
                    ierEntry(this, new EntryEventArgs(EntryType.Join, e.Identity.Nickname, channel));
				}
			}
		}

		private void Chatter_OnLeave(object o, JoinLeaveEventArgs e)
		{
			foreach (IrcString channel in e.GetChannelList())
			{
				ChannelMessage(Color.Green, channel, _["{0} left {1}", e.Identity.Nickname, channel], false);
				ChannelListMessage(channel, new[] { e.Identity.Nickname }, true);

                if (ierEntry != null)
				{
                    ierEntry(this, new EntryEventArgs(EntryType.Leave, e.Identity.Nickname, channel));
				}
			}
		}

        private void Chatter_OnKick(object sender, KickEventArgs e)
        {
            ChannelMessage(Color.Green, e.Channel, e.Reason != null
                ? _["{0} kicked {1} from {2} ({3})", e.Sender.Nickname, e.Recipient, e.Channel, e.Reason]
                : _["{0} kicked {1} from {2}", e.Sender.Nickname, e.Recipient, e.Channel], false);
            ChannelListMessage(e.Channel, new[] { e.Recipient }, true);

            if (ierEntry != null)
            {
                ierEntry(this, new EntryEventArgs(EntryType.Kick, e.Recipient, e.Channel));
            }
        }

        private void Chatter_OnQuit(object sender, QuitEventArgs e)
        {
            foreach (var channelStruct in channels)
            {
                var channel = channelStruct.Channel;
                if (!channel.StartsWith("#")) { continue; }

                ChannelMessage(Color.Green, channel, e.QuitMessage != null
                    ? _["{0} quit ({1})", e.Identity.Nickname, e.QuitMessage]
                    : _["{0} quit", e.Identity.Nickname], false);
                ChannelListMessage(channel, new[] { e.Identity.Nickname }, true);

                if (ierEntry != null)
                {
                    ierEntry(this, new EntryEventArgs(EntryType.Quit, e.Identity.Nickname, channel));
                }
            }
        }

		private void Chatter_OnNameListReply(object o, NameListReplyEventArgs e)
		{
			ChannelMessage(Color.White, e.Channel, null, false);
			ChannelListMessage(e.Channel, e.GetNameList(), false);
		}

		private void MessageAction(Color messageColor, string formatMessage,
			ChatMessageEventArgs e, bool Notice)
		{
            var sender = e.Sender != null ? e.Sender.Nickname : IrcString.Empty;
			ChannelMessage(messageColor, e.Recipient.StartsWith("#") ?
				e.Recipient : sender, _[formatMessage, sender, e.Message],
				Notice);
		}

		private string GetOriginalModeString(ModeEventArgs e)
		{
			string original = e.Command;

			foreach (string s in e.GetParameterList()) { original += " " + s; }

			return original;
		}

		private void Chatter_OnMode(object o, ModeEventArgs e)
		{
			ChannelMessage(Color.DarkGreen, e.Recipient.StartsWith("#") ? e.Recipient : e.Sender.Nickname,
                _["mode change by {0} to {1} of {2}", e.Sender.Nickname, e.Recipient, GetOriginalModeString(e)],
                true);

            bool add = e.Command[0] == '+' ? true : false; int param = 0;

            if (e.Recipient.StartsWith("#"))
            {
                // Channel mode.
                ChannelArray ca = FindChannelWindow(e.Recipient);

                if (ca != null)
                {
                    Channel channel = (Channel)ca.Form;

                    for (int i = 1; i < e.Command.Length; i++)
                    {
                        switch ((char)e.Command[i])
                        {
                            case 'o':
                                {
                                    string name = e.GetParameter(param++);

                                    if (add) { channel.NameChangeRawNew(name, "@" + name); }
                                    else { channel.NameChangeRawBoth("@" + name, name); }
                                }
                                break;

                            case 'p':
                                break;

                            case 's':
                                break;

                            case 'i':
                                break;

                            case 't':
                                break;

                            case 'n':
                                break;

                            case 'm':
                                break;

                            case 'l':
                                {
                                    int limit = int.Parse(e.GetParameter(param++));
                                }
                                break;

                            case 'b':
                                {
                                    string banMask = e.GetParameter(param++);
                                }
                                break;

                            case 'v':
                                {
                                    string name = e.GetParameter(param++);

                                    if (add) { channel.NameChangeRawBoth(name, "+" + name); }
                                    else { channel.NameChangeRawBoth("+" + name, name); }
                                }
                                break;

                            case 'k':
                                {
                                    string key = e.GetParameter(param++);
                                }
                                break;
                        }
                    }
                }
            }
		}

		private void InvokePluginMessage(ChatMessageEventArgs message,
			                             MessageType messageType, MessageFlags messageFlags)
		{
            var sender = message.Sender != null ? message.Sender.Nickname : IrcString.Empty;
			InvokePluginMessage(sender, message.Recipient,
				                message.Message, messageType, messageFlags);
		}

		internal void InvokePluginMessage(string sender, string recipient, string message,
			                              MessageType messageType, MessageFlags messageFlags)
		{
			if (ierMessage != null)
			{
				ierMessage(this, new Plugins.Events.MessageEventArgs
					(sender, recipient, message, messageType, messageFlags));
			}
		}

		private void Chatter_OnMessage(object o, ChatMessageEventArgs e)
		{
			MessageAction(Color.Black, "{0} message: {1}", e, false);
            InvokePluginMessage(e, MessageType.Message, MessageFlags.None);
		}

		private void Chatter_OnChatAction(object o, ChatMessageEventArgs e)
		{
			MessageAction(Color.Purple, "{0} action: {1}", e, false);
            InvokePluginMessage(e, MessageType.Action, MessageFlags.None);
		}

		private void Chatter_OnNotice(object o, ChatMessageEventArgs e)
		{
			MessageAction(Color.Red, "{0} notice: {1}", e, true);
            InvokePluginMessage(e, MessageType.Notice, MessageFlags.None);
		}

		private void Chatter_OnPingReply(object o, PingReplyEventArgs e)
		{
			ChannelMessage(Color.BlueViolet, StatusWindowName,
				_["ping to {0} took {1} ms", e.Identity.Nickname, e.Delay],
                true);
		}

		private void Chatter_OnMOTDText(object o, SimpleMessageEventArgs e)
		{
			ChannelMessage(Color.DarkRed, StatusWindowName,
				_["motd: {0}", e.Message], true); 
		}

		private void Chatter_OnNameChange(object o, NameChangeEventArgs e)
		{
            IrcString OldName = e.Identity.Nickname, NewName = e.NewName;

			bool displayed = false;

			string text = _["name change {0} to {1}", OldName, NewName];

			if (OldName == trueNickname) { trueNickname = NewName; }

			foreach (ChannelArray ca in channels)
			{
				if (ca.Form.NameChange(OldName, NewName))
				{
					displayed = true;

					ChannelMessage(Color.DarkGreen, ca.Channel, text, true);
				}
			}

			if (!displayed)
				{ ChannelMessage(Color.DarkGreen, StatusWindowName, text, true); }
		}

		private void Chatter_OnError(object o, IrcErrorEventArgs e)
		{
			switch (e.Error)
			{
				case IrcReplyCode.NicknameCollision:
				case IrcReplyCode.NicknameInUse:
					ircClient.ChangeName(Settings.Instance.UserInfo.Nickname +
						new Random().Next(10000, 99999).ToString());
					break;
			}

			ChannelMessage(Color.DarkKhaki, StatusWindowName, "Error: " +
				                 e.Error.ToString(), true);
		}

		private void Chatter_OnChannelListBegin(object o, EventArgs e)
		{
			ChannelMessage(Color.DarkKhaki, StatusWindowName, "Channel listing started", true);

			if (chanList != null) { chanList.Close(); }

			chanList = new ChannelList(this);
			chanList.MdiParent = this;
			chanList.Show();
		}


		private void Chatter_OnChannelListEnd(object o, EventArgs e)
		{
			chanList.Done();

			ChannelMessage(Color.DarkKhaki, StatusWindowName, "Channel listing complete", true);
		}

		private void Chatter_OnChannelListEntry(object o, ChannelListEntryEventArgs e)
		{
            if (chanList != null && chanList.Visible)
				{ chanList.AddChannel(e.Channel, e.UserCount, e.Topic); }
		}

		private void mnuCommandsFiltersEliz_Click(object sender, System.EventArgs e)
		{
			mnuCommandsFiltersEliz.Checked = Settings.Instance.Language.Filters.Elizabethan =
				!Settings.Instance.Language.Filters.Elizabethan;
		}

		private void mnuCommandsFiltersPropEng_Click(object sender, System.EventArgs e)
		{
			mnuCommandsFiltersPropEng.Checked = Settings.Instance.Language.Filters.ProperEnglish =
				!Settings.Instance.Language.Filters.ProperEnglish;
		}

		private void mnuCommandsFiltersShort_Click(object sender, System.EventArgs e)
		{
			mnuCommandsFiltersShort.Checked = Settings.Instance.Language.Filters.Shorthand =
				!Settings.Instance.Language.Filters.Shorthand;
		}

		internal string ComputeFilteredMessage(string Message)
		{
			Message = Message.Trim();

			if (Settings.Instance.Language.Filters.Elizabethan)
			{ Message = Filters.ComputeElizabethan(Message); }

			if (Settings.Instance.Language.Filters.ProperEnglish)
			{ Message = Filters.ComputeProperEnglish(Message); }

			if (Settings.Instance.Language.Filters.Shorthand)
			{ Message = Filters.ComputeShorthand(Message); }

			return Message;
		}

		private void menuCommandsList_Click(object sender, System.EventArgs e)
		{
			ircClient.ListChannels();
		}

		private void mnuHelpAbout_Click(object sender, System.EventArgs e)
		{
			new About().ShowDialog();
		}

		private void mnuCommandsLeave_Click(object sender, System.EventArgs e)
		{
			Form form = ActiveMdiChild;
            
			if (form != null)
			{
                var channel = form as Channel;
                if (channel != null && !channel.IsStatusWindow())
                {
                    channel.Close();
                }
			}
		}

		private void mnuCommandsNickname_Click(object sender, System.EventArgs e)
		{
			string name;
			DialogResult result = InputDialog.QueryDialog
				("Change to what nickname?", "Change nickname", s => IrcValidation.IsNickname(s), out name);

			if (result == DialogResult.OK)
				{ ircClient.ChangeName(name); }
		}

		private void mnuWindowCascade_Click(object sender, System.EventArgs e)
		{
			LayoutMdi(MdiLayout.Cascade);
		}

		private void mnuWindowTileH_Click(object sender, System.EventArgs e)
		{
			LayoutMdi(MdiLayout.TileHorizontal);
		}

		private void mnuWindowTileV_Click(object sender, System.EventArgs e)
		{
			LayoutMdi(MdiLayout.TileVertical);
		}

		private void mnuWindowArrange_Click(object sender, System.EventArgs e)
		{
			LayoutMdi(MdiLayout.ArrangeIcons);
		}
	}

    public class FontList
    {
        public Font Channels;
        public Font ChatText;
        public Font UserList;
    }
}